import os
from typing import Any

import yaml

from .base import Package
from .clone import Clone
from .github import GitHubPackage
from .pipx import PipxPackage
from .system import SystemPackage
from .url import URLPackage


def parse_package(package: Any) -> Package:
    if not isinstance(package, dict):
        raise ValueError(f"Dictionary expected: {package}.")
    if "name" in package:
        return SystemPackage(**package)
    elif "repo" in package:
        return GitHubPackage(**package)
    elif "url" in package:
        return URLPackage(**package)
    elif "clone" in package:
        return Clone(**package)
    elif "pipx" in package:
        return PipxPackage(**package)
    else:
        raise ValueError(
            f"Either 'name', 'repo' 'url', 'clone' or 'pipx' must be present, got: {package}."
        )


def load_packages(component: str) -> list[Package]:
    with open(os.path.join("packages", f"{component}.yaml")) as f:
        packages = yaml.safe_load(f)
        return list(map(parse_package, packages))



import os
import tempfile
from abc import ABCMeta, abstractmethod
from typing import Optional, Union

from ..fs import make_executable, makedirs
from ..utils import *
from .manual import ManualPackage


class ArchivePackage(ManualPackage, metaclass=ABCMeta):
    def __init__(
        self,
        *,
        raw: Union[bool, str] = False,
        raw_executable: bool = False,
        strip: int = 0,
        **kwargs,
    ) -> None:
        self.raw = raw
        self.raw_executable = raw_executable
        self.strip = strip
        super().__init__(**kwargs)

    @abstractmethod
    def archive_url(self) -> str:
        pass

    def package_directory(self) -> str:
        result = local(f"{self.name.replace('/', '--')}.app")
        makedirs(result)
        return result

    def untar(self, source: str, *extra: str) -> None:
        run(
            "tar",
            "-x",
            "--strip",
            str(self.strip),
            "-C",
            self.package_directory(),
            *extra,
            "-f",
            source,
        )

    def extract(self, url: str, source: str) -> None:
        if self.raw:
            if isinstance(self.raw, str):
                filename = self.raw
            else:
                filename = url.rsplit("/", 1)[-1]
            target = os.path.join(self.package_directory(), filename)
            run("cp", source, target)
            if self.raw_executable:
                make_executable(target)
        elif ".tar" in url:
            if ".gz" in url:
                self.untar(source, "-z")
            elif ".bz2" in url:
                self.untar(source, "-j")
            else:
                self.untar(source)
        elif ".tgz" in url:
            self.untar(source, "-z")
        elif ".txz" in url:
            self.untar(source, "-J")
        else:
            raise ValueError(f"Unknown archive format: {url}")

    def binary_path(self, binary: str) -> str:
        paths: list[list[str]] = [[], ["bin"]]
        for relative_path in paths:
            candidate = os.path.join(self.package_directory(), *relative_path, binary)
            if os.path.isfile(candidate) and os.access(candidate, os.X_OK):
                return candidate
        raise ValueError(f"Cannot find {binary} in {self.package_directory()}.")

    def app_path(self, name: str) -> str:
        candidate = os.path.join(
            self.package_directory(), "share", "applications", f"{name}.desktop"
        )
        if os.path.isfile(candidate):
            return candidate
        raise ValueError(
            f"Cannot find application '{name}' in {self.package_directory()}."
        )

    def icon_directory(self) -> Optional[str]:
        candidate = os.path.join(self.package_directory(), "share", "icons")
        if os.path.isdir(candidate):
            return candidate
        return None

    def font_path(self, name: str) -> str:
        candidate = os.path.join(self.package_directory(), name)
        if os.path.isfile(candidate):
            return candidate
        raise ValueError(f"Cannot find font '{name}' in {self.package_directory()}.")

    def install(self):
        url = self.archive_url()
        with tempfile.NamedTemporaryFile() as archive_file:
            run("curl", "-sSL", url, stdout=archive_file)
            self.extract(url, archive_file.name)
        super().install()



import os
from pathlib import Path
from typing import Iterator, Literal, Optional

from .utils import run


def home(*path: str) -> str:
    return os.path.join(os.environ["HOME"], *path)


def local(*path: str) -> str:
    return home(".local", *path)


def files_in_recursively(directory: str, glob: str = "*") -> Iterator[str]:
    for path in Path(directory).rglob(glob):
        yield str(path)


def transplant_path(dir_from: str, dir_to: str, path: str) -> str:
    path = os.path.relpath(path, dir_from)
    return os.path.join(dir_to, path)


def makedirs(path: str, sudo: bool = False) -> None:
    if sudo:
        run("sudo", "mkdir", "-p", path)
    else:
        os.makedirs(path, exist_ok=True)


def rm(path: str, sudo: bool = False) -> None:
    if sudo:
        run("sudo", "rm", "-r", "-f", path)
    else:
        if os.path.lexists(path):
            os.unlink(path)


def make_executable(path: str) -> None:
    run("chmod", "+x", path)


LinkMethod = Literal["binary_wrapper"]


def link(
    source: str, target: str, *, sudo: bool = False, method: Optional[LinkMethod] = None
) -> None:
    target_dir = os.path.dirname(target)
    makedirs(target_dir, sudo=sudo)
    rm(target, sudo=sudo)
    if method == "binary_wrapper":
        if sudo:
            raise NotImplementedError()
        else:
            with open(target, "w") as wrapper_file:
                print(f'#!/bin/sh\nexec "{source}" "$@"', file=wrapper_file)
            make_executable(target)
    else:
        if sudo:
            run("sudo", "ln", "-s", "-f", "-T", source, target)
        else:
            os.symlink(source, target)



import os
from typing import Iterator

import yaml

from .fs import files_in_recursively, link
from .utils import *


class Links:
    def __init__(
        self,
        source: str,
        dest: str,
        dot: bool = False,
        root: bool = False,
        shallow: bool = False,
        only: Some[str] = None,
    ) -> None:
        self.root = root
        self.source = os.path.abspath(source)
        if not dest.startswith("/"):
            if self.root:
                dest = os.path.join(with_os(linux="/root", macos="/var/root"), dest)
            else:
                dest = os.path.join(home(), dest)
        self.dest = dest
        self.dot = dot
        self.shallow = shallow
        self.only = unsome_(only)

    def install(self) -> None:
        paths: Iterator[str]
        if self.shallow:
            paths = map(lambda entry: entry.path, os.scandir(self.source))
        else:
            paths = filter(os.path.isfile, files_in_recursively(self.source))

        for path in paths:
            if self.only and os.path.basename(path) not in self.only:
                continue

            target = os.path.relpath(path, self.source)
            if self.dot:
                target = "." + target
            target = os.path.join(self.dest, target)

            link(path, target, sudo=self.root)


def load_links() -> list[Links]:
    with open("links.yaml") as f:
        links = yaml.safe_load(f)
        return list(Links(**args) for args in links)



import argparse

from .dotfiles import load_links
from .package import load_packages
from .utils import flatten, parallel_map_tqdm


def parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Install the environment")
    parser.add_argument("component", nargs="*", help="The component to install")
    return parser


def main():
    args = parser().parse_args()
    components: frozenset[str] = frozenset(args.component) | {"base"}
    packages = flatten(map(load_packages, components))
    parallel_map_tqdm(packages, lambda p: p.ensure())
    links = load_links()
    for link in links:
        link.install()



import configparser
import os
import shutil
from abc import ABCMeta, abstractmethod
from typing import Optional

from ..fs import files_in_recursively, link, makedirs, transplant_path
from ..state import VERSIONS, Version
from ..utils import *
from .base import Package


def icon_name(app_path: str) -> Optional[str]:
    app = configparser.ConfigParser()
    app.read(app_path)
    return app["Desktop Entry"].get("Icon")


class ManualPackage(Package, metaclass=ABCMeta):
    binaries: list[str]
    apps: list[str]
    fonts: list[str]

    def __init__(
        self,
        *,
        as_global: bool = False,
        binary: Some[str] = None,
        binary_wrapper: bool = False,
        app: Some[str] = None,
        font: Some[str] = None,
        **kwargs,
    ) -> None:
        self.as_global = as_global
        self.binaries = unsome(binary)
        self.binary_wrapper = binary_wrapper
        self.apps = unsome(app)
        self.fonts = unsome(font)
        super().__init__(**kwargs)

    @property
    def local_version(self) -> Optional[str]:
        try:
            return VERSIONS[self.name].version
        except KeyError:
            return None

    def local(self, *path: str) -> str:
        if self.as_global:
            return os.path.join("/usr/local", *path)
        else:
            return local(*path)

    @abstractmethod
    def binary_path(self, binary: str) -> str:
        pass

    def install_binary(self, name: str) -> None:
        link(
            self.binary_path(name),
            self.local("bin", name),
            sudo=self.as_global,
            method="binary_wrapper" if self.binary_wrapper else None,
        )

    @abstractmethod
    def app_path(self, name: str) -> str:
        pass

    def icon_directory(self) -> Optional[str]:
        return None

    def install_app(self, name: str) -> None:
        with_os(linux=self.install_app_linux, macos=self.install_app_macos)(name)

    def install_app_linux(self, name: str) -> None:
        path = self.app_path(name)
        target = self.local("share", "applications", f"{name}.desktop")
        link(path, target, sudo=self.as_global)
        icons_source = self.icon_directory()
        if icons_source:
            icons_target = self.local("share", "icons")
            icon = icon_name(path)
            if icon:
                for icon_path in files_in_recursively(icons_source, f"{icon}.*"):
                    target = transplant_path(icons_source, icons_target, icon_path)
                    link(icon_path, target, sudo=self.as_global)

    def install_app_macos(self, name: str) -> None:
        # FIXME: copy to /Applications and/or ~/Applications; ensure names,
        # etc. are correct
        pass

    @abstractmethod
    def font_path(self, name: str) -> str:
        pass

    def install_font(self, name: str) -> None:
        font_dir = with_os(
            linux=self.local("share", "fonts"), macos=home("Library", "Fonts")
        )
        makedirs(font_dir)
        source = self.font_path(name)
        target = os.path.join(font_dir, name)
        link(source, target, sudo=self.as_global)
        if shutil.which("fc-cache"):
            run("fc-cache", "-f", font_dir)

    def install(self) -> None:
        for binary in self.binaries:
            self.install_binary(binary)
        for app in self.apps:
            self.install_app(app)
        for font in self.fonts:
            self.install_font(font)
        VERSIONS[self.name] = Version(version=self.get_remote_version())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mybox', 'mybox.package', 'mybox.state']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'requests>=2.28.1,<3.0.0', 'tqdm>=4.64.0,<5.0.0']

entry_points = \
{'console_scripts': ['mybox = mybox.main:main']}

setup_kwargs = {
    'name': 'mybox',
    'version': '0.0.4',
    'description': 'Manage the configuration and tools on your workstation without bothering the OS too much',
    'long_description': "# Mybox\n\n🖥️ This is a box. 📦 And it is mine. 🐱\n\nThere are many 🍱 nice things in there. I wouldn't want 🧰 to be without them.\n\nEven if I move 🏠 or work 🏢 I want to be comfortable.\n\n---\n\nManage the configuration and tools on your workstation without bothering the OS\ntoo much (maybe your favorite one isn't supported by `$WORK` or you have\ndifferent ones for different roles).\n\n## Status\n\nIn development, for the current horrible state of affairs see\n[install](https://github.com/koterpillar/desktop/blob/main/install).\n",
    'author': 'Alexey Kotlyarov',
    'author_email': 'a@koterpillar.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/koterpillar/mybox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
